import { merge } from "lodash";
import {
  FC,
  ReactNode,
  createContext,
  useContext,
  useEffect,
  useState,
} from "react";
import { defaultSettings, SettingsType } from "../lib/defaultSettings";
import { logcat } from "../lib/logcatService";

// settings context
type SettingsContextProps = {
  settings: SettingsType;
  fetchUserSettingsLocal: () => Promise<void>;
  fetchUserSettingsCloud: () => Promise<void>;
  updateSettingsLocal: (settings: SettingsType) => void;
  updateSettingsCloud: (settings: SettingsType) => Promise<void>;
  deleteSettingsLocal: () => void;
  deleteSettingsCloud: () => Promise<void>;
  settingsLoading: boolean;
};

const SettingsContext = createContext<SettingsContextProps | undefined>(
  undefined,
);

export const SettingsProvider: FC<{ children: ReactNode }> = ({ children }) => {
  logcat.log("Initializing settings...", "INFO");

  // useStates
  const [settings, setSettings] = useState<SettingsType>(defaultSettings);
  const [settingsLoading, setSettingsLoading] = useState<boolean>(true);

  // fetch user settings from localStorage
  const fetchUserSettingsLocal = async () => {
    try {
      logcat.log("Fetching user settings from localStorage...", "INFO");
      const userSettings = localStorage.getItem("userSettings");

      if (userSettings) {
        // deep merge user settings with default settings
        const mergedSettings = merge(
          {},
          defaultSettings,
          JSON.parse(userSettings),
        );

        setSettings(mergedSettings);

        logcat.log("User settings fetched from localStorage", "INFO");
        logcat.log("User settings: " + userSettings, "DEBUG");
      } else {
        logcat.log(
          "User settings not found in localStorage, using default settings",
          "WARN",
        );
      }
      setSettingsLoading(false);
    } catch (error) {
      logcat.log("Error fetching user settings from localStorage", "ERROR");
      logcat.log(String(error), "ERROR");
      setSettingsLoading(false);
    }
  };

  // fetch user settings from Firestore
  const fetchUserSettingsCloud = async () => {
    try {
      if (user) {
        logcat.log("Fetching user settings from firestore...", "INFO");
        const userSettings = await firestoreRetrieveDocumentField(
          `users/${user.uid}`,
          "settings",
        );

        if (userSettings) {
          // deep merge user settings with default settings
          const mergedSettings = merge({}, defaultSettings, userSettings);

          setSettings(mergedSettings);

          logcat.log("User settings fetched from firestore", "INFO");
          logcat.log("User settings: " + JSON.stringify(userSettings), "DEBUG");

          // and then save to localStorage
          updateSettingsLocal(mergedSettings);
        } else {
          logcat.log(
            "User settings not found in firestore, using default settings",
            "WARN",
          );
        }
        setSettingsLoading(false);
      } else {
        logcat.log("User not logged in, using default settings", "WARN");
        setSettingsLoading(false);
      }
    } catch (error) {
      logcat.log("No such field", "ERROR");
      logcat.log(String(error), "ERROR");
      setSettingsLoading(false);
    }
  };

  // push new settings to localStorage
  const updateSettingsLocal = (settings: SettingsType) => {
    try {
      // apply settings into state
      logcat.log("Applying settings...", "INFO");
      setSettings(settings);
      logcat.log("Settings applied", "INFO");

      // save settings to localStorage
      logcat.log("Saving user settings to localStorage...", "INFO");
      localStorage.setItem("userSettings", JSON.stringify(settings));
      logcat.log("User settings saved to localStorage", "INFO");
    } catch (error) {
      logcat.log("Error saving user settings to localStorage", "ERROR");
      logcat.log(String(error), "ERROR");
    }
  };

  // push new settings to firestore
  const updateSettingsCloud = async (settings: SettingsType) => {
    try {
      // apply settings into state
      logcat.log("Applying settings...", "INFO");
      setSettings(settings);
      logcat.log("Settings applied", "INFO");

      // save settings to firestore
      logcat.log("Saving user settings to Firestore...", "INFO");

      // push entire settings object to firestore
      if (user) {
        firestoreCommitUpdate(`users/${user.uid}`, "settings", settings);
        await firestorePushUpdates();

        logcat.log("User settings saved to firestore", "INFO");
      }
    } catch (error) {
      logcat.log("Error saving user settings to firestore", "ERROR");
      logcat.log(String(error), "ERROR");
    }
  };

  // delete settings from localStorage
  const deleteSettingsLocal = () => {
    try {
      logcat.log("Deleting user settings from localStorage...", "INFO");
      localStorage.removeItem("userSettings");
      logcat.log("User settings deleted from localStorage", "INFO");
    } catch (error) {
      logcat.log("Error deleting user settings from localStorage", "ERROR");
      logcat.log(String(error), "ERROR");
    }
  };

  // delete settings from cloud
  const deleteSettingsCloud = async () => {
    try {
      logcat.log("Deleting user settings from Firestore...", "INFO");

      if (user) {
        firestoreCommitDelete(`users/${user.uid}`, "settings");
        await firestorePushDeletes();

        logcat.log("User settings deleted from Firestore", "INFO");
      }
    } catch (error) {
      logcat.log(String(error), "ERROR");
      throw new Error("Error deleting user settings from Firestore");
    }
  };

  // fetch user settings from local on first load everytime
  useEffect(() => {
    fetchUserSettingsLocal();
  }, []);

  return (
    <SettingsContext.Provider
      value={{
        settings,
        fetchUserSettingsLocal,
        fetchUserSettingsCloud,
        updateSettingsLocal,
        updateSettingsCloud,
        deleteSettingsLocal,
        deleteSettingsCloud,
        settingsLoading,
      }}
    >
      {children}
    </SettingsContext.Provider>
  );
};

export const useSettings = () => {
  const context = useContext(SettingsContext);
  if (context === undefined) {
    throw new Error("Please use useSettings only within a SettingsProvider");
  }
  return context;
};
